from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass
from functools import partial as _partial
from typing import TYPE_CHECKING, Any, TypeVar

from typing_extensions import override

from utilities.iterables import one
from utilities.text import ensure_str

if TYPE_CHECKING:
    from collections.abc import Iterator

_T = TypeVar("_T")


class partial(_partial[_T]):  # noqa: N801
    """Partial which accepts Ellipsis for positional arguments."""

    @override
    def __call__(self, *args: Any, **kwargs: Any) -> _T:
        iter_args = iter(args)
        head = (next(iter_args) if arg is ... else arg for arg in self.args)
        return self.func(*head, *iter_args, **{**self.keywords, **kwargs})


@contextmanager
def redirect_empty_reduce() -> Iterator[None]:
    """Redirect to the `EmptyReduceError`."""
    try:
        yield
    except TypeError as error:
        arg = ensure_str(one(error.args))
        if arg == "reduce() of empty iterable with no initial value":
            raise EmptyReduceError from None
        raise


@dataclass(kw_only=True)
class EmptyReduceError(Exception):
    @override
    def __str__(self) -> str:
        return "reduce() must not be called over an empty iterable, or must have an initial value."


__all__ = ["EmptyReduceError", "partial", "redirect_empty_reduce"]



from __future__ import annotations

from hypothesis import given
from hypothesis.strategies import lists
from pytest import mark, param, raises

from utilities.hypothesis import text_ascii
from utilities.sentinel import sentinel
from utilities.text import (
    EnsureStrError,
    ensure_str,
    join_strs,
    split_str,
    strip_and_dedent,
)


class TestEnsureStr:
    @mark.parametrize(
        ("obj", "nullable"), [param("", False), param("", True), param(None, True)]
    )
    def test_main(self, *, obj: bool | None, nullable: bool) -> None:
        _ = ensure_str(obj, nullable=nullable)

    @mark.parametrize(
        ("nullable", "match"),
        [
            param(False, "Object .* must be a string"),
            param(True, "Object .* must be a string or None"),
        ],
    )
    def test_error(self, *, nullable: bool, match: str) -> None:
        with raises(EnsureStrError, match=f"{match}; got .* instead"):
            _ = ensure_str(sentinel, nullable=nullable)


class TestSplitStrAndJoinStr:
    @mark.parametrize(
        ("text", "texts"),
        [
            param("", [""]),
            param("1", ["1"]),
            param("1,2", ["1", "2"]),
            param(",", ["", ""]),
            param(str(sentinel), []),
        ],
    )
    def test_main(self, *, text: str, texts: list[str]) -> None:
        assert split_str(text) == texts
        assert join_strs(texts) == text

    @given(texts=lists(text_ascii()))
    def test_generic(self, *, texts: list[str]) -> None:
        assert split_str(join_strs(texts)) == texts


class TestStripAndDedent:
    def test_main(self) -> None:
        text = """
               This is line 1.
               This is line 2.
               """
        result = strip_and_dedent(text)
        expected = "This is line 1.\nThis is line 2."
        assert result == expected



from __future__ import annotations

from collections import Counter
from collections.abc import Hashable, Iterable, Iterator, Mapping, Sequence, Sized
from collections.abc import Set as AbstractSet
from dataclasses import dataclass
from functools import partial
from itertools import islice, product
from typing import Any, Generic, Literal, TypeGuard, TypeVar, cast, overload

from typing_extensions import Never, assert_never, override

from utilities.math import (
    _CheckIntegerEqualError,
    _CheckIntegerEqualOrApproxError,
    _CheckIntegerMaxError,
    _CheckIntegerMinError,
    check_integer,
)
from utilities.text import ensure_str
from utilities.types import ensure_hashable

_K = TypeVar("_K")
_T = TypeVar("_T")
_V = TypeVar("_V")
_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")
_T3 = TypeVar("_T3")
_T4 = TypeVar("_T4")
_T5 = TypeVar("_T5")
_THashable = TypeVar("_THashable", bound=Hashable)


def check_bijection(mapping: Mapping[Any, Hashable], /) -> None:
    """Check if a mapping is a bijection."""
    try:
        check_duplicates(mapping.values())
    except CheckDuplicatesError as error:
        raise CheckBijectionError(mapping=mapping, counts=error.counts) from None


@dataclass(kw_only=True)
class CheckBijectionError(Exception, Generic[_THashable]):
    mapping: Mapping[Any, _THashable]
    counts: Mapping[_THashable, int]

    @override
    def __str__(self) -> str:
        return "Mapping {} must be a bijection; got duplicates {}.".format(
            self.mapping, ", ".join(f"({k}, n={v})" for k, v in self.counts.items())
        )


def check_duplicates(iterable: Iterable[Hashable], /) -> None:
    """Check if an iterable contains any duplicates."""
    counts = {k: v for k, v in Counter(iterable).items() if v > 1}
    if len(counts) >= 1:
        raise CheckDuplicatesError(iterable=iterable, counts=counts)


@dataclass(kw_only=True)
class CheckDuplicatesError(Exception, Generic[_THashable]):
    iterable: Iterable[_THashable]
    counts: Mapping[_THashable, int]

    @override
    def __str__(self) -> str:
        return (
            f"Iterable {self.iterable} must not contain duplicates; got {self.counts}."
        )


def check_iterables_equal(left: Iterable[Any], right: Iterable[Any], /) -> None:
    """Check that a pair of iterables are equal."""

    left_list, right_list = map(list, [left, right])
    errors: list[tuple[int, Any, Any]] = []
    state: _CheckIterablesEqualState | None
    it = zip(left_list, right_list, strict=True)
    try:
        for i, (lv, rv) in enumerate(it):
            if lv != rv:
                errors.append((i, lv, rv))
    except ValueError as error:
        msg = ensure_str(one(error.args))
        match msg:
            case "zip() argument 2 is longer than argument 1":
                state = "right_longer"
            case "zip() argument 2 is shorter than argument 1":
                state = "left_longer"
            case _:  # pragma: no cover
                raise
    else:
        state = None
    if (len(errors) >= 1) or (state is not None):
        raise CheckIterablesEqualError(
            left=left_list, right=right_list, errors=errors, state=state
        )


_CheckIterablesEqualState = Literal["left_longer", "right_longer"]


@dataclass(kw_only=True)
class CheckIterablesEqualError(Exception, Generic[_T]):
    left: list[_T]
    right: list[_T]
    errors: list[tuple[int, _T, _T]]
    state: _CheckIterablesEqualState | None

    @override
    def __str__(self) -> str:
        match list(self._yield_parts()):
            case (desc,):
                pass
            case first, second:
                desc = f"{first} and {second}"
            case _ as never:  # pragma: no cover
                assert_never(cast(Never, never))
        return f"Iterables {self.left} and {self.right} must be equal; {desc}."

    def _yield_parts(self) -> Iterator[str]:
        if len(self.errors) >= 1:
            error_descs = (f"({lv}, {rv}, i={i})" for i, lv, rv in self.errors)
            yield "differing items were {}".format(", ".join(error_descs))
        match self.state:
            case "left_longer":
                yield "left was longer"
            case "right_longer":
                yield "right was longer"
            case None:
                pass
            case _ as never:  # type: ignore[]
                assert_never(never)


def check_length(
    obj: Sized,
    /,
    *,
    equal: int | None = None,
    equal_or_approx: int | tuple[int, float] | None = None,
    min: int | None = None,  # noqa: A002
    max: int | None = None,  # noqa: A002
) -> None:
    """Check the length of an object."""
    n = len(obj)
    try:
        check_integer(n, equal=equal, equal_or_approx=equal_or_approx, min=min, max=max)
    except _CheckIntegerEqualError as error:
        raise _CheckLengthEqualError(obj=obj, equal=error.equal) from None
    except _CheckIntegerEqualOrApproxError as error:
        raise _CheckLengthEqualOrApproxError(
            obj=obj, equal_or_approx=error.equal_or_approx
        ) from None
    except _CheckIntegerMinError as error:
        raise _CheckLengthMinError(obj=obj, min_=error.min_) from None
    except _CheckIntegerMaxError as error:
        raise _CheckLengthMaxError(obj=obj, max_=error.max_) from None


@dataclass(kw_only=True)
class CheckLengthError(Exception):
    obj: Sized


@dataclass(kw_only=True)
class _CheckLengthEqualError(CheckLengthError):
    equal: int

    @override
    def __str__(self) -> str:
        return f"Object {self.obj} must have length {self.equal}; got {len(self.obj)}."


@dataclass(kw_only=True)
class _CheckLengthEqualOrApproxError(CheckLengthError):
    equal_or_approx: int | tuple[int, float]

    @override
    def __str__(self) -> str:
        match self.equal_or_approx:
            case target, error:
                desc = f"approximate length {target} (error {error:%})"
            case target:
                desc = f"length {target}"
        return f"Object {self.obj} must have {desc}; got {len(self.obj)}."


@dataclass(kw_only=True)
class _CheckLengthMinError(CheckLengthError):
    min_: int

    @override
    def __str__(self) -> str:
        return f"Object {self.obj} must have minimum length {self.min_}; got {len(self.obj)}."


@dataclass(kw_only=True)
class _CheckLengthMaxError(CheckLengthError):
    max_: int

    @override
    def __str__(self) -> str:
        return f"Object {self.obj} must have maximum length {self.max_}; got {len(self.obj)}."


def check_lengths_equal(left: Sized, right: Sized, /) -> None:
    """Check that a pair of sizes objects have equal length."""
    if len(left) != len(right):
        raise CheckLengthsEqualError(left=left, right=right)


@dataclass(kw_only=True)
class CheckLengthsEqualError(Exception):
    left: Sized
    right: Sized

    @override
    def __str__(self) -> str:
        return f"Sized objects {self.left} and {self.right} must have the same length; got {len(self.left)} and {len(self.right)}."


def check_mappings_equal(left: Mapping[Any, Any], right: Mapping[Any, Any], /) -> None:
    """Check that a pair of mappings are equal."""
    left_keys, right_keys = set(left), set(right)
    try:
        check_sets_equal(left_keys, right_keys)
    except CheckSetsEqualError as error:
        left_extra, right_extra = map(set, [error.left_extra, error.right_extra])
    else:
        left_extra = right_extra = set()
    errors: list[tuple[Any, Any, Any]] = []
    for key in left_keys & right_keys:
        lv, rv = left[key], right[key]
        if lv != rv:
            errors.append((key, lv, rv))
    if (len(left_extra) >= 1) or (len(right_extra) >= 1) or (len(errors) >= 1):
        raise CheckMappingsEqualError(
            left=left,
            right=right,
            left_extra=left_extra,
            right_extra=right_extra,
            errors=errors,
        )


@dataclass(kw_only=True)
class CheckMappingsEqualError(Exception, Generic[_K, _V]):
    left: Mapping[_K, _V]
    right: Mapping[_K, _V]
    left_extra: AbstractSet[_K]
    right_extra: AbstractSet[_K]
    errors: list[tuple[_K, _V, _V]]

    @override
    def __str__(self) -> str:
        match list(self._yield_parts()):
            case (desc,):
                pass
            case first, second:
                desc = f"{first} and {second}"
            case first, second, third:
                desc = f"{first}, {second} and {third}"
            case _ as never:  # pragma: no cover
                assert_never(cast(Never, never))
        return f"Mappings {self.left} and {self.right} must be equal; {desc}."

    def _yield_parts(self) -> Iterator[str]:
        if len(self.left_extra) >= 1:
            yield f"left had extra keys {self.left_extra}"
        if len(self.right_extra) >= 1:
            yield f"right had extra keys {self.right_extra}"
        if len(self.errors) >= 1:
            error_descs = (f"({lv}, {rv}, k={k})" for k, lv, rv in self.errors)
            yield "differing values were {}".format(", ".join(error_descs))


def check_sets_equal(left: Iterable[Any], right: Iterable[Any], /) -> None:
    """Check that a pair of sets are equal."""
    left_as_set = set(left)
    right_as_set = set(right)
    left_extra = left_as_set - right_as_set
    right_extra = right_as_set - left_as_set
    if (len(left_extra) >= 1) or (len(right_extra) >= 1):
        raise CheckSetsEqualError(
            left=left_as_set,
            right=right_as_set,
            left_extra=left_extra,
            right_extra=right_extra,
        )


@dataclass(kw_only=True)
class CheckSetsEqualError(Exception, Generic[_T]):
    left: AbstractSet[_T]
    right: AbstractSet[_T]
    left_extra: AbstractSet[_T]
    right_extra: AbstractSet[_T]

    @override
    def __str__(self) -> str:
        match list(self._yield_parts()):
            case (desc,):
                pass
            case first, second:
                desc = f"{first} and {second}"
            case _ as never:  # pragma: no cover
                assert_never(cast(Never, never))
        return f"Sets {self.left} and {self.right} must be equal; {desc}."

    def _yield_parts(self) -> Iterator[str]:
        if len(self.left_extra) >= 1:
            yield f"left had extra items {self.left_extra}"
        if len(self.right_extra) >= 1:
            yield f"right had extra items {self.right_extra}"


def check_submapping(left: Mapping[Any, Any], right: Mapping[Any, Any], /) -> None:
    """Check that a mapping is a subset of another mapping."""
    left_keys, right_keys = set(left), set(right)
    try:
        check_subset(left_keys, right_keys)
    except CheckSubSetError as error:
        extra = set(error.extra)
    else:
        extra = set()
    errors: list[tuple[Any, Any, Any]] = []
    for key in left_keys & right_keys:
        lv, rv = left[key], right[key]
        if lv != rv:
            errors.append((key, lv, rv))
    if (len(extra) >= 1) or (len(errors) >= 1):
        raise CheckSubMappingError(left=left, right=right, extra=extra, errors=errors)


@dataclass(kw_only=True)
class CheckSubMappingError(Exception, Generic[_K, _V]):
    left: Mapping[_K, _V]
    right: Mapping[_K, _V]
    extra: AbstractSet[_K]
    errors: list[tuple[_K, _V, _V]]

    @override
    def __str__(self) -> str:
        match list(self._yield_parts()):
            case (desc,):
                pass
            case first, second:
                desc = f"{first} and {second}"
            case _ as never:  # pragma: no cover
                assert_never(cast(Never, never))
        return f"Mapping {self.left} must be a submapping of {self.right}; {desc}."

    def _yield_parts(self) -> Iterator[str]:
        if len(self.extra) >= 1:
            yield f"left had extra keys {self.extra}"
        if len(self.errors) >= 1:
            error_descs = (f"({lv}, {rv}, k={k})" for k, lv, rv in self.errors)
            yield "differing values were {}".format(", ".join(error_descs))


def check_subset(left: Iterable[Any], right: Iterable[Any], /) -> None:
    """Check that a set is a subset of another set."""
    left_as_set = set(left)
    right_as_set = set(right)
    extra = left_as_set - right_as_set
    if len(extra) >= 1:
        raise CheckSubSetError(left=left_as_set, right=right_as_set, extra=extra)


@dataclass(kw_only=True)
class CheckSubSetError(Exception, Generic[_T]):
    left: AbstractSet[_T]
    right: AbstractSet[_T]
    extra: AbstractSet[_T]

    @override
    def __str__(self) -> str:
        return f"Set {self.left} must be a subset of {self.right}; left had extra items {self.extra}."


def check_supermapping(left: Mapping[Any, Any], right: Mapping[Any, Any], /) -> None:
    """Check that a mapping is a superset of another mapping."""
    left_keys, right_keys = set(left), set(right)
    try:
        check_superset(left_keys, right_keys)
    except CheckSuperSetError as error:
        extra = set(error.extra)
    else:
        extra = set()
    errors: list[tuple[Any, Any, Any]] = []
    for key in left_keys & right_keys:
        lv, rv = left[key], right[key]
        if lv != rv:
            errors.append((key, lv, rv))
    if (len(extra) >= 1) or (len(errors) >= 1):
        raise CheckSuperMappingError(left=left, right=right, extra=extra, errors=errors)


@dataclass(kw_only=True)
class CheckSuperMappingError(Exception, Generic[_K, _V]):
    left: Mapping[_K, _V]
    right: Mapping[_K, _V]
    extra: AbstractSet[_K]
    errors: list[tuple[_K, _V, _V]]

    @override
    def __str__(self) -> str:
        match list(self._yield_parts()):
            case (desc,):
                pass
            case first, second:
                desc = f"{first} and {second}"
            case _ as never:  # pragma: no cover
                assert_never(cast(Never, never))
        return f"Mapping {self.left} must be a supermapping of {self.right}; {desc}."

    def _yield_parts(self) -> Iterator[str]:
        if len(self.extra) >= 1:
            yield f"right had extra keys {self.extra}"
        if len(self.errors) >= 1:
            error_descs = (f"({lv}, {rv}, k={k})" for k, lv, rv in self.errors)
            yield "differing values were {}".format(", ".join(error_descs))


def check_superset(left: Iterable[Any], right: Iterable[Any], /) -> None:
    """Check that a set is a superset of another set."""
    left_as_set = set(left)
    right_as_set = set(right)
    extra = right_as_set - left_as_set
    if len(extra) >= 1:
        raise CheckSuperSetError(left=left_as_set, right=right_as_set, extra=extra)


@dataclass(kw_only=True)
class CheckSuperSetError(Exception, Generic[_T]):
    left: AbstractSet[_T]
    right: AbstractSet[_T]
    extra: AbstractSet[_T]

    @override
    def __str__(self) -> str:
        return f"Set {self.left} must be a superset of {self.right}; right had extra items {self.extra}."


def chunked(iterable: Iterable[_T], n: int, /) -> Iterable[Sequence[_T]]:
    """Break an iterable into lists of length n."""
    return cast(Iterable[Sequence[_T]], iter(partial(take, n, iter(iterable)), []))


def ensure_hashables(
    *args: Any, **kwargs: Any
) -> tuple[list[Hashable], dict[str, Hashable]]:
    """Ensure a set of positional & keyword arguments are all hashable."""
    hash_args = list(map(ensure_hashable, args))
    hash_kwargs = {k: ensure_hashable(v) for k, v in kwargs.items()}
    return hash_args, hash_kwargs


def ensure_iterable(obj: Any, /) -> Iterable[Any]:
    """Ensure an object is iterable."""
    if is_iterable(obj):
        return obj
    raise EnsureIterableError(obj=obj)


@dataclass(kw_only=True)
class EnsureIterableError(Exception):
    obj: Any

    @override
    def __str__(self) -> str:
        return f"Object {self.obj} must be iterable."


def ensure_iterable_not_str(obj: Any, /) -> Iterable[Any]:
    """Ensure an object is iterable, but not a string."""
    if is_iterable_not_str(obj):
        return obj
    raise EnsureIterableNotStrError(obj=obj)


@dataclass(kw_only=True)
class EnsureIterableNotStrError(Exception):
    obj: Any

    @override
    def __str__(self) -> str:
        return f"Object {self.obj} must be iterable, but not a string."


def is_iterable(obj: Any, /) -> TypeGuard[Iterable[Any]]:
    """Check if an object is iterable."""
    try:
        iter(obj)
    except TypeError:
        return False
    return True


def is_iterable_not_str(obj: Any, /) -> TypeGuard[Iterable[Any]]:
    """Check if an object is iterable, but not a string."""
    return is_iterable(obj) and not isinstance(obj, str)


def one(iterable: Iterable[_T], /) -> _T:
    """Custom version of `one` with separate exceptions."""
    it = iter(iterable)
    try:
        first = next(it)
    except StopIteration:
        raise OneEmptyError(iterable=iterable) from None
    try:
        second = next(it)
    except StopIteration:
        return first
    raise OneNonUniqueError(iterable=iterable, first=first, second=second)


@dataclass(kw_only=True)
class OneError(Exception, Generic[_T]):
    iterable: Iterable[_T]


@dataclass(kw_only=True)
class OneEmptyError(OneError[_T]):
    @override
    def __str__(self) -> str:
        return f"Iterable {self.iterable} must not be empty."


@dataclass(kw_only=True)
class OneNonUniqueError(OneError[_T]):
    first: _T
    second: _T

    @override
    def __str__(self) -> str:
        return f"Iterable {self.iterable} must contain exactly one item; got {self.first}, {self.second} and perhaps more."


def one_str(
    iterable: Iterable[str], text: str, /, *, case_sensitive: bool = True
) -> str:
    """Find the unique string in an iterable."""
    as_list = list(iterable)
    try:
        check_duplicates(as_list)
    except CheckDuplicatesError as error:
        raise _OneStrDuplicatesError(
            iterable=iterable, text=text, counts=error.counts
        ) from None
    if case_sensitive:
        try:
            return one(t for t in as_list if t == text)
        except OneEmptyError:
            raise _OneStrCaseSensitiveEmptyError(iterable=iterable, text=text) from None
    mapping = {t: t.casefold() for t in as_list}
    try:
        check_bijection(mapping)
    except CheckBijectionError as error:
        error = cast(CheckBijectionError[str], error)
        raise _OneStrCaseInsensitiveBijectionError(
            iterable=iterable, text=text, counts=error.counts
        ) from None
    try:
        return one(k for k, v in mapping.items() if v == text.casefold())
    except OneEmptyError:
        raise _OneStrCaseInsensitiveEmptyError(iterable=iterable, text=text) from None


@dataclass(kw_only=True)
class OneStrError(Exception):
    iterable: Iterable[str]
    text: str


@dataclass(kw_only=True)
class _OneStrDuplicatesError(OneStrError):
    counts: Mapping[str, int]

    @override
    def __str__(self) -> str:
        return (
            f"Iterable {self.iterable} must not contain duplicates; got {self.counts}."
        )


@dataclass(kw_only=True)
class _OneStrCaseSensitiveEmptyError(OneStrError):
    @override
    def __str__(self) -> str:
        return f"Iterable {self.iterable} does not contain {self.text!r}."


@dataclass(kw_only=True)
class _OneStrCaseInsensitiveBijectionError(OneStrError):
    counts: Mapping[str, int]

    @override
    def __str__(self) -> str:
        return f"Iterable {self.iterable} must not contain duplicates (case insensitive); got {self.counts}."


@dataclass(kw_only=True)
class _OneStrCaseInsensitiveEmptyError(OneStrError):
    @override
    def __str__(self) -> str:
        return f"Iterable {self.iterable} does not contain {self.text!r} (case insensitive)."


def product_dicts(mapping: Mapping[_K, Iterable[_V]], /) -> Iterable[Mapping[_K, _V]]:
    """The cartesian product of the values in a mapping, as mappings."""
    keys = list(mapping)
    for values in product(*mapping.values()):
        yield dict(zip(keys, values, strict=True))


def take(n: int, iterable: Iterable[_T], /) -> Sequence[_T]:
    """Return first n items of the iterable as a list."""
    return list(islice(iterable, n))


@overload
def transpose(iterable: Iterable[tuple[_T1]], /) -> tuple[tuple[_T1, ...]]: ...
@overload
def transpose(
    iterable: Iterable[tuple[_T1, _T2]], /
) -> tuple[tuple[_T1, ...], tuple[_T2, ...]]: ...
@overload
def transpose(
    iterable: Iterable[tuple[_T1, _T2, _T3]], /
) -> tuple[tuple[_T1, ...], tuple[_T2, ...], tuple[_T3, ...]]: ...
@overload
def transpose(
    iterable: Iterable[tuple[_T1, _T2, _T3, _T4]], /
) -> tuple[tuple[_T1, ...], tuple[_T2, ...], tuple[_T3, ...], tuple[_T4, ...]]: ...
@overload
def transpose(
    iterable: Iterable[tuple[_T1, _T2, _T3, _T4, _T5]], /
) -> tuple[
    tuple[_T1, ...], tuple[_T2, ...], tuple[_T3, ...], tuple[_T4, ...], tuple[_T5, ...]
]: ...
def transpose(iterable: Iterable[tuple[Any, ...]]) -> tuple[tuple[Any, ...], ...]:
    """Typed verison of `transpose`."""
    return tuple(zip(*iterable, strict=True))


__all__ = [
    "CheckBijectionError",
    "CheckDuplicatesError",
    "CheckIterablesEqualError",
    "CheckLengthsEqualError",
    "CheckMappingsEqualError",
    "CheckSetsEqualError",
    "CheckSubMappingError",
    "CheckSubSetError",
    "CheckSuperMappingError",
    "CheckSuperSetError",
    "EnsureIterableError",
    "EnsureIterableNotStrError",
    "OneEmptyError",
    "OneError",
    "OneNonUniqueError",
    "check_bijection",
    "check_duplicates",
    "check_iterables_equal",
    "check_lengths_equal",
    "check_mappings_equal",
    "check_sets_equal",
    "check_submapping",
    "check_subset",
    "check_supermapping",
    "check_superset",
    "chunked",
    "ensure_hashables",
    "ensure_iterable",
    "ensure_iterable_not_str",
    "is_iterable",
    "is_iterable_not_str",
    "one",
    "product_dicts",
    "take",
    "transpose",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, cast, overload

import cvxpy
import numpy as np
import numpy.linalg
from cvxpy import CLARABEL, Expression, Problem
from numpy import ndarray, where
from typing_extensions import override

from utilities.numpy import NDArrayF, NDArrayF1, NDArrayF2, is_non_zero, is_zero

if TYPE_CHECKING:
    from pandas import DataFrame

    from utilities.pandas import SeriesF


@overload
def abs_(x: float, /) -> float: ...
@overload
def abs_(x: NDArrayF, /) -> NDArrayF: ...
@overload
def abs_(x: SeriesF, /) -> SeriesF: ...
@overload
def abs_(x: DataFrame, /) -> DataFrame: ...
@overload
def abs_(x: Expression, /) -> Expression: ...
def abs_(
    x: float | NDArrayF | SeriesF | DataFrame | Expression, /
) -> float | NDArrayF | SeriesF | DataFrame | Expression:
    """Compute the absolute value."""
    if isinstance(x, int | float | ndarray):
        return np.abs(x)
    try:
        from pandas import DataFrame, Series
    except ModuleNotFoundError:  # pragma: no cover
        pass
    else:
        if isinstance(x, Series | DataFrame):
            return np.abs(x)
    return cvxpy.abs(x)


@overload
def add(x: float, y: float, /) -> float: ...
@overload
def add(x: NDArrayF, y: float, /) -> NDArrayF: ...
@overload
def add(x: Expression, y: float, /) -> Expression: ...
@overload
def add(x: float, y: NDArrayF, /) -> NDArrayF: ...
@overload
def add(x: NDArrayF, y: NDArrayF, /) -> NDArrayF: ...
@overload
def add(x: Expression, y: NDArrayF, /) -> Expression: ...
@overload
def add(x: float, y: Expression, /) -> Expression: ...
@overload
def add(x: NDArrayF, y: Expression, /) -> Expression: ...
@overload
def add(x: Expression, y: Expression, /) -> Expression: ...
def add(
    x: float | NDArrayF | Expression, y: float | NDArrayF | Expression, /
) -> float | NDArrayF | Expression:
    """Compute the sum of two quantities."""
    if isinstance(x, int | float | ndarray) and isinstance(y, int | float | ndarray):
        return np.add(x, y)
    return cast(Any, x) + cast(Any, y)


@overload
def divide(x: float, y: float, /) -> float: ...
@overload
def divide(x: NDArrayF, y: float, /) -> NDArrayF: ...
@overload
def divide(x: Expression, y: float, /) -> Expression: ...
@overload
def divide(x: float, y: NDArrayF, /) -> NDArrayF: ...
@overload
def divide(x: NDArrayF, y: NDArrayF, /) -> NDArrayF: ...
@overload
def divide(x: Expression, y: NDArrayF, /) -> Expression: ...
@overload
def divide(x: float, y: Expression, /) -> Expression: ...
@overload
def divide(x: NDArrayF, y: Expression, /) -> Expression: ...
@overload
def divide(x: Expression, y: Expression, /) -> Expression: ...
def divide(
    x: float | NDArrayF | Expression, y: float | NDArrayF | Expression, /
) -> float | NDArrayF | Expression:
    """Compute the quotient of two quantities."""
    if isinstance(x, int | float | ndarray) and isinstance(y, int | float | ndarray):
        return np.divide(x, y)
    return cast(Any, x) / cast(Any, y)


@overload
def max_(x: float | NDArrayF | SeriesF | DataFrame, /) -> float: ...
@overload
def max_(x: Expression, /) -> Expression: ...


def max_(
    x: float | NDArrayF | SeriesF | DataFrame | Expression, /
) -> float | Expression:
    """Compute the maximum of a quantity."""
    if isinstance(x, int | float | ndarray):
        return np.max(x)
    try:
        from pandas import DataFrame, Series
    except ModuleNotFoundError:  # pragma: no cover
        pass
    else:
        if isinstance(x, Series | DataFrame):
            return max_(x.to_numpy())
    return cvxpy.max(x)


@overload
def maximum(x: float, y: float, /) -> float: ...
@overload
def maximum(x: NDArrayF, y: float, /) -> NDArrayF: ...
@overload
def maximum(x: SeriesF, y: float, /) -> SeriesF: ...
@overload
def maximum(x: DataFrame, y: float, /) -> DataFrame: ...
@overload
def maximum(x: Expression, y: float, /) -> Expression: ...
@overload
def maximum(x: float, y: NDArrayF, /) -> NDArrayF: ...
@overload
def maximum(x: NDArrayF, y: NDArrayF, /) -> NDArrayF: ...
@overload
def maximum(x: SeriesF, y: NDArrayF, /) -> SeriesF: ...
@overload
def maximum(x: DataFrame, y: NDArrayF, /) -> DataFrame: ...
@overload
def maximum(x: Expression, y: NDArrayF, /) -> Expression: ...
@overload
def maximum(x: float, y: SeriesF, /) -> SeriesF: ...
@overload
def maximum(x: NDArrayF, y: SeriesF, /) -> SeriesF: ...
@overload
def maximum(x: SeriesF, y: SeriesF, /) -> SeriesF: ...
@overload
def maximum(x: Expression, y: SeriesF, /) -> Expression: ...
@overload
def maximum(x: float, y: DataFrame, /) -> DataFrame: ...
@overload
def maximum(x: NDArrayF, y: DataFrame, /) -> DataFrame: ...
@overload
def maximum(x: DataFrame, y: DataFrame, /) -> DataFrame: ...
@overload
def maximum(x: Expression, y: DataFrame, /) -> Expression: ...
@overload
def maximum(x: float, y: Expression, /) -> Expression: ...
@overload
def maximum(x: NDArrayF, y: Expression, /) -> Expression: ...
@overload
def maximum(x: SeriesF, y: Expression, /) -> Expression: ...
@overload
def maximum(x: DataFrame, y: Expression, /) -> Expression: ...
@overload
def maximum(x: Expression, y: Expression, /) -> Expression: ...
def maximum(
    x: float | NDArrayF | SeriesF | DataFrame | Expression,
    y: float | NDArrayF | SeriesF | DataFrame | Expression,
    /,
) -> float | NDArrayF | SeriesF | DataFrame | Expression:
    """Compute the elementwise maximum of two quantities."""
    try:
        _check_series_and_dataframe(x, y)
    except _CheckSeriesAndDataFrameError as error:
        raise MaximumError(x=error.x, y=error.y) from None
    try:
        from pandas import DataFrame, Series
    except ModuleNotFoundError:  # pragma: no cover
        if isinstance(x, int | float | ndarray) and isinstance(
            y, int | float | ndarray
        ):
            return np.maximum(x, y)
    else:
        if isinstance(x, int | float | ndarray | Series | DataFrame) and isinstance(
            y, int | float | ndarray | Series | DataFrame
        ):
            return np.maximum(x, y)
    return cvxpy.maximum(x, y)


@dataclass(kw_only=True)
class MaximumError(Exception):
    x: SeriesF | DataFrame
    y: SeriesF | DataFrame

    @override
    def __str__(self) -> str:
        return f"Maximum must not be between a Series and DataFrame; got {self.x} and {self.y}."


@overload
def min_(x: float | NDArrayF | SeriesF | DataFrame, /) -> float: ...
@overload
def min_(x: Expression, /) -> Expression: ...
def min_(
    x: float | NDArrayF | SeriesF | DataFrame | Expression, /
) -> float | Expression:
    """Compute the minimum of a quantity."""
    if isinstance(x, int | float | ndarray):
        return np.min(x)
    try:
        from pandas import DataFrame, Series
    except ModuleNotFoundError:  # pragma: no cover
        pass
    else:
        if isinstance(x, Series | DataFrame):
            return min_(x.to_numpy())
    return cvxpy.min(x)


@overload
def minimum(x: float, y: float, /) -> float: ...
@overload
def minimum(x: NDArrayF, y: float, /) -> NDArrayF: ...
@overload
def minimum(x: SeriesF, y: float, /) -> SeriesF: ...
@overload
def minimum(x: DataFrame, y: float, /) -> DataFrame: ...
@overload
def minimum(x: Expression, y: float, /) -> Expression: ...
@overload
def minimum(x: float, y: NDArrayF, /) -> NDArrayF: ...
@overload
def minimum(x: NDArrayF, y: NDArrayF, /) -> NDArrayF: ...
@overload
def minimum(x: SeriesF, y: NDArrayF, /) -> SeriesF: ...
@overload
def minimum(x: DataFrame, y: NDArrayF, /) -> DataFrame: ...
@overload
def minimum(x: Expression, y: NDArrayF, /) -> Expression: ...
@overload
def minimum(x: float, y: SeriesF, /) -> SeriesF: ...
@overload
def minimum(x: NDArrayF, y: SeriesF, /) -> SeriesF: ...
@overload
def minimum(x: SeriesF, y: SeriesF, /) -> SeriesF: ...
@overload
def minimum(x: Expression, y: SeriesF, /) -> Expression: ...
@overload
def minimum(x: float, y: DataFrame, /) -> DataFrame: ...
@overload
def minimum(x: NDArrayF, y: DataFrame, /) -> DataFrame: ...
@overload
def minimum(x: DataFrame, y: DataFrame, /) -> DataFrame: ...
@overload
def minimum(x: Expression, y: DataFrame, /) -> Expression: ...
@overload
def minimum(x: float, y: Expression, /) -> Expression: ...
@overload
def minimum(x: NDArrayF, y: Expression, /) -> Expression: ...
@overload
def minimum(x: SeriesF, y: Expression, /) -> Expression: ...
@overload
def minimum(x: DataFrame, y: Expression, /) -> Expression: ...
@overload
def minimum(x: Expression, y: Expression, /) -> Expression: ...
def minimum(
    x: float | NDArrayF | SeriesF | DataFrame | Expression,
    y: float | NDArrayF | SeriesF | DataFrame | Expression,
    /,
) -> float | NDArrayF | SeriesF | DataFrame | Expression:
    """Compute the elementwise minimum of two quantities."""
    try:
        _check_series_and_dataframe(x, y)
    except _CheckSeriesAndDataFrameError as error:
        raise MinimumError(x=error.x, y=error.y) from None
    try:
        from pandas import DataFrame, Series
    except ModuleNotFoundError:  # pragma: no cover
        if isinstance(x, int | float | ndarray) and isinstance(
            y, int | float | ndarray
        ):
            return np.minimum(x, y)
    else:
        if isinstance(x, int | float | ndarray | Series | DataFrame) and isinstance(
            y, int | float | ndarray | Series | DataFrame
        ):
            return np.minimum(x, y)
    return cvxpy.minimum(x, y)


@dataclass(kw_only=True)
class MinimumError(Exception):
    x: SeriesF | DataFrame
    y: SeriesF | DataFrame

    @override
    def __str__(self) -> str:
        return f"Minimum must not be between a Series and DataFrame; got {self.x} and {self.y}."


@overload
def multiply(x: float, y: float, /) -> float: ...
@overload
def multiply(x: NDArrayF, y: float, /) -> NDArrayF: ...
@overload
def multiply(x: SeriesF, y: float, /) -> SeriesF: ...
@overload
def multiply(x: DataFrame, y: float, /) -> DataFrame: ...
@overload
def multiply(x: Expression, y: float, /) -> Expression: ...
@overload
def multiply(x: float, y: NDArrayF, /) -> NDArrayF: ...
@overload
def multiply(x: NDArrayF, y: NDArrayF, /) -> NDArrayF: ...
@overload
def multiply(x: SeriesF, y: NDArrayF, /) -> SeriesF: ...
@overload
def multiply(x: DataFrame, y: NDArrayF, /) -> DataFrame: ...
@overload
def multiply(x: Expression, y: NDArrayF, /) -> Expression: ...
@overload
def multiply(x: float, y: SeriesF, /) -> SeriesF: ...
@overload
def multiply(x: NDArrayF, y: SeriesF, /) -> SeriesF: ...
@overload
def multiply(x: SeriesF, y: SeriesF, /) -> SeriesF: ...
@overload
def multiply(x: Expression, y: SeriesF, /) -> Expression: ...
@overload
def multiply(x: float, y: DataFrame, /) -> DataFrame: ...
@overload
def multiply(x: NDArrayF, y: DataFrame, /) -> DataFrame: ...
@overload
def multiply(x: DataFrame, y: DataFrame, /) -> DataFrame: ...
@overload
def multiply(x: Expression, y: DataFrame, /) -> Expression: ...
@overload
def multiply(x: float, y: Expression, /) -> Expression: ...
@overload
def multiply(x: NDArrayF, y: Expression, /) -> Expression: ...
@overload
def multiply(x: SeriesF, y: Expression, /) -> Expression: ...
@overload
def multiply(x: DataFrame, y: Expression, /) -> Expression: ...
@overload
def multiply(x: Expression, y: Expression, /) -> Expression: ...
def multiply(
    x: float | NDArrayF | SeriesF | DataFrame | Expression,
    y: float | NDArrayF | SeriesF | DataFrame | Expression,
    /,
) -> float | NDArrayF | SeriesF | DataFrame | Expression:
    """Compute the elementwise product of two quantities."""
    try:
        _check_series_and_dataframe(x, y)
    except _CheckSeriesAndDataFrameError as error:
        raise MultiplyError(x=error.x, y=error.y) from None
    try:
        from pandas import DataFrame, Series
    except ModuleNotFoundError:  # pragma: no cover
        if isinstance(x, int | float | ndarray) and isinstance(
            y, int | float | ndarray
        ):
            return np.multiply(x, y)
    else:
        if isinstance(x, int | float | ndarray | Series | DataFrame) and isinstance(
            y, int | float | ndarray | Series | DataFrame
        ):
            return np.multiply(x, y)
    return cvxpy.multiply(x, y)


@dataclass(kw_only=True)
class MultiplyError(Exception):
    x: SeriesF | DataFrame
    y: SeriesF | DataFrame

    @override
    def __str__(self) -> str:
        return f"Multiply must not be between a Series and DataFrame; got {self.x} and {self.y}."


@overload
def negate(x: float, /) -> float: ...
@overload
def negate(x: NDArrayF, /) -> NDArrayF: ...
@overload
def negate(x: SeriesF, /) -> SeriesF: ...
@overload
def negate(x: DataFrame, /) -> DataFrame: ...
@overload
def negate(x: Expression, /) -> Expression: ...
def negate(
    x: float | NDArrayF | SeriesF | DataFrame | Expression, /
) -> float | NDArrayF | SeriesF | DataFrame | Expression:
    """Negate a quantity."""
    return -x


@overload
def negative(x: float, /) -> float: ...
@overload
def negative(x: NDArrayF, /) -> NDArrayF: ...
@overload
def negative(x: SeriesF, /) -> SeriesF: ...
@overload
def negative(x: DataFrame, /) -> DataFrame: ...
@overload
def negative(x: Expression, /) -> Expression: ...
def negative(
    x: float | NDArrayF | SeriesF | DataFrame | Expression, /
) -> float | NDArrayF | SeriesF | DataFrame | Expression:
    """Compute the negative parts of a quantity."""
    if isinstance(x, int | float | ndarray):
        result = -minimum(x, 0.0)
        return where(is_zero(result), 0.0, result)
    try:
        from pandas import DataFrame, Series
    except ModuleNotFoundError:  # pragma: no cover
        pass
    else:
        if isinstance(x, Series | DataFrame):
            result = -minimum(x, 0.0)
            return result.where(is_non_zero(result), 0.0)
    return cvxpy.neg(x)


@overload
def norm(x: NDArrayF1 | SeriesF, /) -> float: ...
@overload
def norm(x: Expression, /) -> Expression: ...
def norm(x: NDArrayF1 | SeriesF | Expression, /) -> float | Expression:
    """Compute the norm of a quantity."""
    if isinstance(x, ndarray):
        return cast(float, numpy.linalg.norm(x))
    try:
        from pandas import Series
    except ModuleNotFoundError:  # pragma: no cover
        pass
    else:
        if isinstance(x, Series):
            return norm(x.to_numpy())
    return cvxpy.norm(x)


@overload
def positive(x: float, /) -> float: ...
@overload
def positive(x: NDArrayF, /) -> NDArrayF: ...
@overload
def positive(x: SeriesF, /) -> SeriesF: ...
@overload
def positive(x: DataFrame, /) -> DataFrame: ...
@overload
def positive(x: Expression, /) -> Expression: ...
def positive(
    x: float | NDArrayF | SeriesF | DataFrame | Expression, /
) -> float | NDArrayF | SeriesF | DataFrame | Expression:
    """Compute the positive parts of a quantity."""
    if isinstance(x, int | float | ndarray):
        result = maximum(x, 0.0)
        return where(is_zero(result), 0.0, result)
    try:
        from pandas import DataFrame, Series
    except ModuleNotFoundError:  # pragma: no cover
        pass
    else:
        if isinstance(x, Series | DataFrame):
            result = maximum(x, 0.0)
            return result.where(is_non_zero(result), 0.0)
    return cvxpy.pos(x)


@overload
def power(x: float, p: float, /) -> float: ...
@overload
def power(x: NDArrayF, p: float, /) -> NDArrayF: ...
@overload
def power(x: Expression, p: float, /) -> Expression: ...
@overload
def power(x: float, p: NDArrayF, /) -> NDArrayF: ...
@overload
def power(x: NDArrayF, p: NDArrayF, /) -> NDArrayF: ...
@overload
def power(x: Expression, p: NDArrayF, /) -> Expression: ...
def power(
    x: float | NDArrayF | Expression, p: float | NDArrayF, /
) -> float | NDArrayF | Expression:
    """Compute the power of a quantity."""
    if isinstance(x, int | float | ndarray):
        return np.power(x, p)
    return cvxpy.power(x, p)


@overload
def quad_form(x: NDArrayF1, P: NDArrayF2, /) -> float: ...  # noqa: N803
@overload
def quad_form(x: Expression, P: NDArrayF2, /) -> Expression: ...  # noqa: N803
def quad_form(
    x: NDArrayF1 | Expression,
    P: NDArrayF2,  # noqa: N803
    /,
) -> float | Expression:
    """Compute the quadratic form of a vector & matrix."""
    if isinstance(x, ndarray):
        return cast(float, x.T @ P @ x)
    return cvxpy.quad_form(x, P)


@overload
def scalar_product(x: float, y: float, /) -> float: ...
@overload
def scalar_product(x: NDArrayF, y: float, /) -> NDArrayF: ...
@overload
def scalar_product(x: SeriesF, y: float, /) -> SeriesF: ...
@overload
def scalar_product(x: DataFrame, y: float, /) -> DataFrame: ...
@overload
def scalar_product(x: Expression, y: float, /) -> Expression: ...
@overload
def scalar_product(x: float, y: NDArrayF, /) -> NDArrayF: ...
@overload
def scalar_product(x: NDArrayF, y: NDArrayF, /) -> NDArrayF: ...
@overload
def scalar_product(x: SeriesF, y: NDArrayF, /) -> SeriesF: ...
@overload
def scalar_product(x: DataFrame, y: NDArrayF, /) -> DataFrame: ...
@overload
def scalar_product(x: Expression, y: NDArrayF, /) -> Expression: ...
@overload
def scalar_product(x: float, y: SeriesF, /) -> SeriesF: ...
@overload
def scalar_product(x: NDArrayF, y: SeriesF, /) -> SeriesF: ...
@overload
def scalar_product(x: SeriesF, y: SeriesF, /) -> SeriesF: ...
@overload
def scalar_product(x: Expression, y: SeriesF, /) -> Expression: ...
@overload
def scalar_product(x: float, y: DataFrame, /) -> DataFrame: ...
@overload
def scalar_product(x: NDArrayF, y: DataFrame, /) -> DataFrame: ...
@overload
def scalar_product(x: DataFrame, y: DataFrame, /) -> DataFrame: ...
@overload
def scalar_product(x: Expression, y: DataFrame, /) -> Expression: ...
@overload
def scalar_product(x: float, y: Expression, /) -> Expression: ...
@overload
def scalar_product(x: NDArrayF, y: Expression, /) -> Expression: ...
@overload
def scalar_product(x: SeriesF, y: Expression, /) -> Expression: ...
@overload
def scalar_product(x: DataFrame, y: Expression, /) -> Expression: ...
@overload
def scalar_product(x: Expression, y: Expression, /) -> Expression: ...
def scalar_product(
    x: float | NDArrayF | SeriesF | DataFrame | Expression,
    y: float | NDArrayF | SeriesF | DataFrame | Expression,
    /,
) -> float | NDArrayF | SeriesF | DataFrame | Expression:
    """Compute the scalar product of two quantities."""
    try:
        prod = multiply(cast(Any, x), cast(Any, y))
    except MultiplyError as error:
        raise ScalarProductError(x=error.x, y=error.y) from None
    return sum_(prod)


@dataclass(kw_only=True)
class ScalarProductError(Exception):
    x: SeriesF | DataFrame
    y: SeriesF | DataFrame

    @override
    def __str__(self) -> str:
        return f"Scalar product must not be between a Series and DataFrame; got {self.x} and {self.y}."


def solve(
    problem: Problem,
    /,
    *,
    solver: Literal[
        "CBC",
        "CLARABEL",
        "COPT",
        "CVXOPT",
        "ECOS",
        "GLOP",
        "GLPK_MI",
        "GLPK",
        "GUROBI",
        "MOSEK",
        "NAG",
        "OSQP",
        "PDLPCPLEX",
        "PIQP",
        "PROXQP",
        "SCIP",
        "SCIPY",
        "SCS",
        "SDPA",
        "XPRESS",
    ] = CLARABEL,
    verbose: bool = False,
    **kwargs: Any,
) -> float:
    """Solve a problem."""
    match solver:
        case "MOSEK":  # pragma: no cover
            specific = {"mosek_params": {"MSK_IPAR_LICENSE_WAIT": True}}
        case _:
            specific = {}
    obj = cast(
        float, problem.solve(solver=solver, verbose=verbose, **kwargs, **specific)
    )
    if (status := problem.status) in {"optimal", "optimal_inaccurate"}:
        return obj
    if status in {"infeasible", "infeasible_inaccurate"}:
        msg = f"{problem=}"
        raise SolveInfeasibleError(msg)
    if status in {"unbounded", "unbounded_inaccurate"}:
        msg = f"{problem=}"
        raise SolveUnboundedError(msg)
    msg = f"{status=}"  # pragma: no cover
    raise SolveError(msg)  # pragma: no cover


class SolveError(Exception): ...


class SolveInfeasibleError(SolveError): ...


class SolveUnboundedError(SolveError): ...


@overload
def sqrt(x: float, /) -> float: ...
@overload
def sqrt(x: NDArrayF, /) -> NDArrayF: ...
@overload
def sqrt(x: SeriesF, /) -> SeriesF: ...
@overload
def sqrt(x: DataFrame, /) -> DataFrame: ...
@overload
def sqrt(x: Expression, /) -> Expression: ...
def sqrt(
    x: float | NDArrayF | SeriesF | DataFrame | Expression, /
) -> float | NDArrayF | SeriesF | DataFrame | Expression:
    """Compute the square root of a quantity."""
    try:
        from pandas import DataFrame, Series
    except ModuleNotFoundError:  # pragma: no cover
        if isinstance(x, int | float | ndarray):
            return np.sqrt(x)
    else:
        if isinstance(x, int | float | ndarray | Series | DataFrame):
            return np.sqrt(x)
    return cvxpy.sqrt(x)


@overload
def subtract(x: float, y: float, /) -> float: ...
@overload
def subtract(x: NDArrayF, y: float, /) -> NDArrayF: ...
@overload
def subtract(x: Expression, y: float, /) -> Expression: ...
@overload
def subtract(x: float, y: NDArrayF, /) -> NDArrayF: ...
@overload
def subtract(x: NDArrayF, y: NDArrayF, /) -> NDArrayF: ...
@overload
def subtract(x: Expression, y: NDArrayF, /) -> Expression: ...
@overload
def subtract(x: float, y: Expression, /) -> Expression: ...
@overload
def subtract(x: NDArrayF, y: Expression, /) -> Expression: ...
@overload
def subtract(x: Expression, y: Expression, /) -> Expression: ...
def subtract(
    x: float | NDArrayF | Expression, y: float | NDArrayF | Expression, /
) -> float | NDArrayF | Expression:
    """Compute the difference of two quantities."""
    if isinstance(x, int | float | ndarray) and isinstance(y, int | float | ndarray):
        return np.subtract(x, y)
    return cast(Any, x) - cast(Any, y)


@overload
def sum_(x: float | NDArrayF | SeriesF | DataFrame, /) -> float: ...
@overload
def sum_(x: Expression, /) -> Expression: ...
def sum_(
    x: float | NDArrayF | SeriesF | DataFrame | Expression, /
) -> float | Expression:
    """Compute the sum of a quantity."""
    if isinstance(x, int | float):
        return x
    if isinstance(x, ndarray):
        return np.sum(x).item()
    try:
        from pandas import DataFrame, Series
    except ModuleNotFoundError:  # pragma: no cover
        pass
    else:
        if isinstance(x, Series | DataFrame):
            return sum_(x.to_numpy())
    return cvxpy.sum(x)


@overload
def sum_axis0(x: NDArrayF2, /) -> NDArrayF1: ...
@overload
def sum_axis0(x: DataFrame, /) -> SeriesF: ...
@overload
def sum_axis0(x: Expression, /) -> Expression: ...
def sum_axis0(
    x: NDArrayF2 | DataFrame | Expression, /
) -> NDArrayF1 | SeriesF | Expression:
    """Compute the sum along axis 0 of a quantity."""
    return _sum_axis_0_or_1(x, 0)


@overload
def sum_axis1(x: NDArrayF2, /) -> NDArrayF1: ...
@overload
def sum_axis1(x: DataFrame, /) -> SeriesF: ...
@overload
def sum_axis1(x: Expression, /) -> Expression: ...
def sum_axis1(
    x: NDArrayF2 | DataFrame | Expression, /
) -> NDArrayF1 | SeriesF | Expression:
    """Compute the sum along axis 1 of a quantity."""
    return _sum_axis_0_or_1(x, 1)


def _sum_axis_0_or_1(
    x: NDArrayF2 | DataFrame | Expression, axis: Literal[0, 1], /
) -> NDArrayF1 | SeriesF | Expression:
    try:
        from pandas import DataFrame
    except ModuleNotFoundError:  # pragma: no cover
        if isinstance(x, ndarray):
            return np.sum(x, axis=axis)
    else:
        if isinstance(x, ndarray | DataFrame):
            return np.sum(x, axis=axis)
    return cast(Expression, cvxpy.sum(x, axis=axis))


def _check_series_and_dataframe(
    x: float | NDArrayF | SeriesF | DataFrame | Expression,
    y: float | NDArrayF | SeriesF | DataFrame | Expression,
    /,
) -> None:
    try:
        from pandas import DataFrame, Series
    except ModuleNotFoundError:  # pragma: no cover
        return
    if (isinstance(x, Series) and isinstance(y, DataFrame)) or (
        isinstance(x, DataFrame) and isinstance(y, Series)
    ):
        raise _CheckSeriesAndDataFrameError(x=x, y=y)


@dataclass(kw_only=True)
class _CheckSeriesAndDataFrameError(Exception):
    x: SeriesF | DataFrame
    y: SeriesF | DataFrame

    @override
    def __str__(self) -> str:
        return f"Function must not be between a Series and DataFrame; got {self.x} and {self.y}."


__all__ = [
    "MaximumError",
    "MinimumError",
    "MultiplyError",
    "ScalarProductError",
    "SolveError",
    "SolveInfeasibleError",
    "SolveUnboundedError",
    "abs_",
    "add",
    "divide",
    "max_",
    "maximum",
    "min_",
    "minimum",
    "multiply",
    "negate",
    "negative",
    "norm",
    "positive",
    "power",
    "quad_form",
    "scalar_product",
    "solve",
    "sqrt",
    "subtract",
    "sum_",
    "sum_axis0",
    "sum_axis1",
]



from __future__ import annotations

from re import DOTALL
from typing import TYPE_CHECKING

from pytest import mark, param, raises

from utilities.re import (
    _ExtractGroupMultipleCaptureGroupsError,
    _ExtractGroupMultipleMatchesError,
    _ExtractGroupNoCaptureGroupsError,
    _ExtractGroupNoMatchesError,
    _ExtractGroupsMultipleMatchesError,
    _ExtractGroupsNoCaptureGroupsError,
    _ExtractGroupsNoMatchesError,
    extract_group,
    extract_groups,
)

if TYPE_CHECKING:
    from utilities.types import IterableStrs


class TestExtractGroup:
    def test_main(self) -> None:
        assert extract_group(r"(\d)", "A0A") == "0"

    def test_with_flags(self) -> None:
        assert extract_group(r"(.\d)", "A\n0\nA", flags=DOTALL) == "\n0"

    @mark.parametrize(
        ("pattern", "text", "error", "match"),
        [
            param(
                r"\d",
                "0",
                _ExtractGroupNoCaptureGroupsError,
                "Pattern .* must contain exactly one capture group; it had none",
            ),
            param(
                r"(\d)(\w)",
                "0A",
                _ExtractGroupMultipleCaptureGroupsError,
                "Pattern .* must contain exactly one capture group; it had multiple",
            ),
            param(
                r"(\d)",
                "A",
                _ExtractGroupNoMatchesError,
                "Pattern .* must match against .*",
            ),
            param(
                r"(\d)",
                "0A0",
                _ExtractGroupMultipleMatchesError,
                "Pattern .* must match against .* exactly once; matches were .*",
            ),
        ],
    )
    def test_errors(
        self, *, pattern: str, text: str, error: type[Exception], match: str
    ) -> None:
        with raises(error, match=match):
            _ = extract_group(pattern, text)


class TestExtractGroups:
    @mark.parametrize(
        ("pattern", "text", "expected"),
        [param(r"(\d)", "A0A", ["0"]), param(r"(\d)(\w)", "A0A0", ["0", "A"])],
    )
    def test_main(self, *, pattern: str, text: str, expected: IterableStrs) -> None:
        assert extract_groups(pattern, text) == expected

    def test_with_flags(self) -> None:
        assert extract_groups(r"(.)(\d)(\w)", "\n0A\n", flags=DOTALL) == [
            "\n",
            "0",
            "A",
        ]

    @mark.parametrize(
        ("pattern", "text", "error", "match"),
        [
            param(
                r"\d",
                "0",
                _ExtractGroupsNoCaptureGroupsError,
                "Pattern .* must contain at least one capture group",
            ),
            param(
                r"(\d)",
                "A",
                _ExtractGroupsNoMatchesError,
                "Pattern .* must match against .*",
            ),
            param(
                r"(\d)",
                "0A0",
                _ExtractGroupsMultipleMatchesError,
                r"Pattern .* must match against .* exactly once; matches were \[.*, .*\]",
            ),
            param(
                r"(\d)(\w)",
                "A0",
                _ExtractGroupsNoMatchesError,
                "Pattern .* must match against .*",
            ),
            param(
                r"(\d)(\w)",
                "0A0A",
                _ExtractGroupsMultipleMatchesError,
                r"Pattern .* must match against .* exactly once; matches were \[.*, .*\]",
            ),
        ],
    )
    def test_errors(
        self, *, pattern: str, text: str, error: type[Exception], match: str
    ) -> None:
        with raises(error, match=match):
            _ = extract_groups(pattern, text)
